@@ -1,5 +1,6 @@
 /*
- * uconv version 0.4.9
+ * uconv version 0.4.10
+ * Sep  4, 2002 yoshidam  fixes memory leaks
  * Dec 10, 2001 yoshidam  supports the tainted status
  * Nov 23, 2001 yoshidam  appends shortest_flag
  * Nov 24, 1999 yoshidam  appends eliminate_zwnbsp_flag
@@ -55,8 +56,10 @@ _u8tou16(unsigned char* in, UString* out)
 	  inlen -= 2;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 31) << 6) | (in[1] & 63);
       UStr_addChar2(out, u & 0xff, u >> 8);
@@ -74,8 +77,10 @@ _u8tou16(unsigned char* in, UString* out)
 	  inlen -= 3;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 15) << 12) | ((in[1] & 63) << 6) | (in[2] & 63);
       /* surrogate chars */
@@ -86,8 +91,10 @@ _u8tou16(unsigned char* in, UString* out)
 	  inlen -= 3;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	}
       }
       UStr_addChar2(out, u & 0xff, u >> 8);
       in += 3;
@@ -105,8 +112,10 @@ _u8tou16(unsigned char* in, UString* out)
 	  inlen -= 4;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 7) << 18) | ((in[1] & 63) << 12) |
 	((in[2] & 63) << 6) | (in[3] & 63);
@@ -124,8 +133,10 @@ _u8tou16(unsigned char* in, UString* out)
 	  inlen -= 4;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	}
       }
       //      UStr_addWCharToU16LE(out, u);
       in += 4;
@@ -137,8 +148,10 @@ _u8tou16(unsigned char* in, UString* out)
 	in++;
 	inlen--;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "illegal UTF-8 sequence (%02x)", c);
+      }
     }
   }
 
@@ -173,8 +186,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 2;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 31) << 6) | (in[1] & 63);
       UStr_addChar4(out, u & 0xff, u >> 8, 0, 0);
@@ -192,8 +207,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 3;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 15) << 12) | ((in[1] & 63) << 6) | (in[2] & 63);
       /* surrogate chars */
@@ -204,8 +221,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 3;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	}
       }
       UStr_addChar4(out, u & 0xff, u >> 8, 0, 0);
       in += 3;
@@ -223,8 +242,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 4;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 7) << 18) | ((in[1] & 63) << 12) |
 	((in[2] & 63) << 6) | (in[3] & 63);
@@ -242,8 +263,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 4;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	}
       }
 #endif /* UTF32 */
       in += 4;
@@ -263,8 +286,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 5;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 3) << 24) | ((in[1] & 63) << 18) |
 	((in[2] & 63) << 12) | ((in[3] & 63) << 6) | (in[4] & 63);
@@ -287,8 +312,10 @@ _u8tou4(unsigned char* in, UString* out)
 	  inlen -= 6;
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "non-shortest UTF-8 sequence (%02x)", c);
+	}
       }
       u = ((c & 1) << 30) | ((in[1] & 63) << 24) |
 	((in[2] & 63) << 18) | ((in[3] & 63) << 12) |
@@ -305,8 +332,10 @@ _u8tou4(unsigned char* in, UString* out)
 	in++;
 	inlen--;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "illegal UTF-8 sequence (%02x)", c);
+      }
     }
   }
 
@@ -340,8 +369,10 @@ _u16tou8(unsigned char* in, int len, UString* out, int eliminate_zwnbsp)
 	UStr_addWChar(out, replace_invalid);
 	continue;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "invalid surrogate detected");
+      }
     }
     else if (c >= 0xd800 && c <= 0xdbff) { /* high surrogate */
       unsigned int low;
@@ -351,8 +382,10 @@ _u16tou8(unsigned char* in, int len, UString* out, int eliminate_zwnbsp)
 	  UStr_addWChar(out, replace_invalid);
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "invalid surrogate detected");
+	}
       }
       low = in[i+2] | (in[i+3] << 8);
       if (low < 0xdc00 || low > 0xdfff) { /* not low surrogate */
@@ -360,8 +393,10 @@ _u16tou8(unsigned char* in, int len, UString* out, int eliminate_zwnbsp)
 	  UStr_addWChar(out, replace_invalid);
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "invalid surrogate detected");
+	}
       }
       c = (((c & 1023)) << 10 | (low & 1023)) + 0x10000;
       b4 = c & 63;
@@ -410,8 +445,10 @@ _u4tou8(unsigned char* in, int len, UString* out, int eliminate_zwnbsp)
 	UStr_addWChar(out, replace_invalid);
 	continue;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "illegal char detected (%04x)", c);
+      }
     }
     else if (c < 0x10000) {     /* 0x0800-0xFFFF */
       unsigned char b3 = c & 63;
@@ -453,8 +490,10 @@ _u4tou8(unsigned char* in, int len, UString* out, int eliminate_zwnbsp)
     else {
       if (replace_invalid)
 	UStr_addWChar(out, replace_invalid);
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "non-UCS char detected");
+      }
     }
   }
 
@@ -479,8 +518,10 @@ _u4tou16(unsigned char* in, int len, UString* out)
 	UStr_addWCharToU16LE(out, replace_invalid);
 	continue;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", c);
+      }
     }
     else if (c < 0x10000) {     /* 0x0000-0xFFFF */
       UStr_addChar2(out, in[i], in[i+1]);
@@ -495,8 +536,10 @@ _u4tou16(unsigned char* in, int len, UString* out)
 	UStr_addWCharToU16LE(out, replace_invalid);
 	continue;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "non-UTF-16 char detected");
+      }
     }
   }
 
@@ -520,8 +563,10 @@ _u16tou4(unsigned char* in, int len, UString* out)
 	UStr_addWCharToU32LE(out, replace_invalid);
 	continue;
       }
-      else
+      else {
+	UStr_free(out);
 	rb_raise(eUconvError, "invalid surrogate detected");
+      }
     }
     else if (c >= 0xd800 && c <= 0xdbff) { /* high surrogate */
       unsigned int low;
@@ -530,8 +575,10 @@ _u16tou4(unsigned char* in, int len, UString* out)
 	  UStr_addWCharToU32LE(out, replace_invalid);
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "invalid surrogate detected");
+	}
       }
       low = in[i+2] | (in[i+3] << 8);
       if (low < 0xdc00 || low > 0xdfff) { /* not low surrogate */
@@ -539,8 +586,10 @@ _u16tou4(unsigned char* in, int len, UString* out)
 	  UStr_addWCharToU32LE(out, replace_invalid);
 	  continue;
 	}
-	else
+	else {
+	  UStr_free(out);
 	  rb_raise(eUconvError, "invalid surrogate detected");
+	}
       }
       c = (((c & 1023)) << 10 | (low & 1023)) + 0x10000;
       UStr_addChar4(out, c & 255, (c >> 8) & 255, c >> 16, 0);
@@ -554,57 +603,76 @@ _u16tou4(unsigned char* in, int len, UString* out)
   return out->len;
 }
 
+static VALUE
+exception_handler(VALUE args, VALUE errinfo) {
+  return errinfo;
+}
+
 #ifdef USE_EUC
-static unsigned short
-unknown_euc_handler(const unsigned char* seq)
+static VALUE
+call_euc_handler(const unsigned char* seq)
 {
   ID mid = rb_intern("unknown_euc_handler");
   VALUE ret;
 
   if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0))
-    return '?';
+    return INT2FIX('?');
 
   ret = rb_funcall((VALUE)mUconv, mid, 1, rb_str_new2((char*)seq));
   Check_Type(ret, T_FIXNUM);
-  return FIX2INT(ret);
+  return ret;
+}
+
+static VALUE
+unknown_euc_handler(const unsigned char* seq)
+{
+  return rb_rescue((VALUE (*)(void*))call_euc_handler, (int)seq,
+		   exception_handler, NULL);
 }
 #endif /* USE_EUC */
 
 #ifdef USE_SJIS
-static unsigned short
-unknown_sjis_handler(const unsigned char* seq)
+static VALUE
+call_sjis_handler(const unsigned char* seq)
 {
   ID mid = rb_intern("unknown_sjis_handler");
   VALUE ret;
 
   if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0))
-    return '?';
+    return INT2FIX('?');
 
   ret = rb_funcall((VALUE)mUconv, mid, 1, rb_str_new2((char*)seq));
   Check_Type(ret, T_FIXNUM);
-  return FIX2INT(ret);
+  return ret;
+}
+
+static VALUE
+unknown_sjis_handler(const unsigned char* seq)
+{
+  return rb_rescue((VALUE (*)(void*))call_sjis_handler, (int)seq,
+		   exception_handler, NULL);
 }
 #endif /* USE_SJIS */
 
-static unsigned char*
-unknown_unicode_handler(unsigned short code)
+static VALUE
+call_unicode_handler(unsigned short code)
 {
   ID mid = rb_intern("unknown_unicode_handler");
   VALUE ret;
-  unsigned char* estr;
 
   if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0)) {
-    estr = (unsigned char*)malloc(2);
-    estr[0] = '?';
-    estr[1] = '\0';
-    return estr;
+    return rb_str_new2("?");
   }
   ret = rb_funcall((VALUE)mUconv, mid, 1, INT2FIX(code));
   Check_Type(ret, T_STRING);
-  estr = (unsigned char*)malloc(RSTRING(ret)->len + 1);
-  memcpy(estr, RSTRING(ret)->ptr, RSTRING(ret)->len + 1);
+  return ret;
+}
 
-  return estr;
+static VALUE
+unknown_unicode_handler(unsigned short code)
+{
+  return rb_rescue((VALUE (*)(void*))call_unicode_handler, code,
+		   exception_handler, NULL);
 }
 
 #ifdef USE_EUC
@@ -915,14 +983,17 @@ uconv_u8toeuc(VALUE obj, VALUE ustr)
   UString out;
   UString e;
   VALUE ret;
+  volatile VALUE vout;
 
   Check_Type(ustr, T_STRING);
   in = (unsigned char*)(RSTRING(ustr)->ptr);
   _u8tou16(in, &out);
-  u2e_conv2(out.str, out.len, &e, unknown_unicode_handler);
+  vout = rb_str_new((char*)out.str, out.len);
+  UStr_free(&out);
+  u2e_conv2(RSTRING(vout)->ptr, RSTRING(vout)->len,
+	    &e, unknown_unicode_handler);
   ret = rb_str_new((char*)e.str, e.len);
   UStr_free(&e);
-  UStr_free(&out);
   OBJ_INFECT(ret, ustr);
 
   return ret;
@@ -936,15 +1007,17 @@ uconv_euctou8(VALUE obj, VALUE estr)
   UString in;
   UString out;
   VALUE ret;
+  VALUE vin;
 
   Check_Type(estr, T_STRING);
   e = (unsigned char*)(RSTRING(estr)->ptr);
   len = RSTRING(estr)->len;
   e2u_conv2(e, &in, unknown_euc_handler);
-  _u16tou8(in.str, in.len, &out, 1);
+  vin = rb_str_new((char*)in.str, in.len);
+  UStr_free(&in);
+  _u16tou8(RSTRING(vin)->ptr, RSTRING(vin)->len, &out, 1);
   ret = rb_str_new((char*)out.str, out.len);
   UStr_free(&out);
-  UStr_free(&in);
   OBJ_INFECT(ret, estr);
 
   return ret;
@@ -959,14 +1032,17 @@ uconv_u8tosjis(VALUE obj, VALUE ustr)
   UString out;
   UString s;
   VALUE ret;
+  volatile VALUE vout;
 
   Check_Type(ustr, T_STRING);
   in = (unsigned char*)(RSTRING(ustr)->ptr);
   _u8tou16(in, &out);
-  u2s_conv2(out.str, out.len, &s, unknown_unicode_handler);
+  vout = rb_str_new((char*)out.str, out.len);
+  UStr_free(&out);
+  u2s_conv2(RSTRING(vout)->ptr, RSTRING(vout)->len,
+	    &s, unknown_unicode_handler);
   ret = rb_str_new((char*)s.str, s.len);
   UStr_free(&s);
-  UStr_free(&out);
   OBJ_INFECT(ret, ustr);
 
   return ret;
@@ -980,15 +1056,17 @@ uconv_sjistou8(VALUE obj, VALUE sstr)
   UString in;
   UString out;
   VALUE ret;
+  VALUE vin;
 
   Check_Type(sstr, T_STRING);
   s = (unsigned char*)(RSTRING(sstr)->ptr);
   len = RSTRING(sstr)->len;
   s2u_conv2(s, &in, unknown_sjis_handler);
-  _u16tou8(in.str, in.len, &out, 1);
+  vin = rb_str_new((char*)in.str, in.len);
+  UStr_free(&in);
+  _u16tou8(RSTRING(vin)->ptr, RSTRING(vin)->len, &out, 1);
   ret = rb_str_new((char*)out.str, out.len);
   UStr_free(&out);
-  UStr_free(&in);
   OBJ_INFECT(ret, sstr);
 
   return ret;
